"""
This module contains all the logic for pwatch

"""
import psutil

from .defaults import defaults


class PWatch:
    """
    The main class for pwatch

    """

    def __init__(self, interval=None, memory_trigger=None, cpu_trigger=None):
        """
        Initialize the process watcher

        """
        from .log import get_logger

        self.logger = get_logger()
        self.run_interval = interval or defaults.run_interval
        self.memory_trigger = memory_trigger
        self.cpu_trigger = cpu_trigger


    def run(self):
        """
        Process all pids and report those who are using excessive resources

        """
        self.logger.debug('Processing...')
        for p in psutil.process_iter():
            if self.check_process(p):
                try:
                    self.report(p)
                except Exception as e:
                    self.logger.warning("Cannot generate report for process {} ({}): {}"
                                        .format(p.name(), p.pid, e))


    def loop(self):
        """
        Run the process watcher forever

        """
        import time

        start_log = 'Started pwatch with limits: '
        if self.cpu_trigger is not None:
            start_log += 'cpu: {:.2f}% '.format(self.cpu_trigger)
        if self.memory_trigger is not None:
            start_log += 'mem: {:.2f} MB '.format(float(self.memory_trigger) / 1024 / 1024)
        self.logger.info(start_log)
        while 42:
            self.run()
            time.sleep(self.run_interval)


    def report(self, process):
        """
        Generate a report for the process and write it to the logger

        """
        import time

        try:
            io_counters = process.io_counters()
        except:
            from collections import namedtuple
            IOCounters = namedtuple("IOCounters", 'read_count write_count read_bytes write_bytes')
            io_counters= IOCounters(read_count=-1, write_count=-1,
                                    read_bytes=0, write_bytes=0)

        r = ("Process {name}({pid}) detected excessive: "
             "mem: {mem:.2f} MB | "
             "cpu: {cpu:.2f}% | "
             "uptime: {uptime}s | "
             "threads: {thr} | "
             "FDs: {fds} | "
             "IO reads: {ior} ({ior_sz:.2f} kB) | "
             "IO writes: {iow} ({iow_sz:.2f} kB) | "
             "context switches: {ctx_switches} | "
             "command: {cmd}"
             ).format(
                name=process.name(),
                pid=process.pid,
                mem=float(process.memory_info().rss) / 1024 / 1024,
                cpu=process.last_cpu_percent,
                uptime=int(time.time() - process.create_time()),
                thr=process.num_threads(),
                fds=process.num_fds(),
                ior=io_counters.read_count,
                iow=io_counters.write_count,
                ior_sz=float(io_counters.read_bytes) / 1024,
                iow_sz=float(io_counters.write_bytes) / 1024,
                ctx_switches=process.num_ctx_switches().voluntary,
                cmd=' '.join(process.cmdline()),
             )
        self.logger.error(r)


    def check_memory(self, process):
        """
        Check if process is above memory usage trigger

        """
        if self.memory_trigger is None:
            return False
        try:
            return process.memory_info().rss >= self.memory_trigger
        except Exception as e:
            self.logger.warning("Error occured while checking memory info for {} ({}): {}"
                                .format(process.name(), process.pid, e))
            return False


    def check_cpu(self, process):
        """
        Check if process is above CPU usage trigger

        """
        if self.cpu_trigger is None:
            return False
        try:
            process.last_cpu_percent = process.cpu_percent()
            return process.last_cpu_percent >= self.cpu_trigger
        except Exception as e:
            self.logger.warning("Error occured while checking CPU info for {} ({}): {}"
                                .format(process.name(), process.pid, e))
            return False


    def check_process(self, process):
        """
        Check a process against memory and cpu triggers

        """
        return self.check_memory(process) or self.check_cpu(process)



"""
This module contains logging utilities for pwatch

"""

_logger = None


def get_logger():
    """
    Returns the logging instance

    """
    import logging
    global _logger

    if _logger is None:
        _logger = logging.getLogger('pwatch')
    return _logger


def setup_logger(debug=False, verbose=False, logger=None):
    """
    Create a logger object and configure it with different handlers

    """
    import logging
    from logging import handlers as lhandlers

    logger = logger or get_logger()
    logger.setLevel(logging.DEBUG if debug else logging.INFO)
    if verbose:
        logger.addHandler(logging.StreamHandler())
    try:
        handler = logging.handlers.SysLogHandler('/dev/log')
        handler.setFormatter(logging.Formatter(
            '%(asctime)s %(name)s: %(message)s',
            '%b %e %H:%M:%S'
        ))
        logger.addHandler(handler)
    except Exception as e:
        logger.error("Cannot attach syslog handler to logger: {}".format(e.strerror))
    return logger



"""
This module contains entry points for pwatch

"""
import sys

from . import pwatch
from .defaults import defaults
from .log import setup_logger


def main():
    """
    Main entry point for pwatch

    """
    import argparse

    parser = argparse.ArgumentParser(
        description='This program is a simple process watcher daemon',
        epilog='Copyright Quanta 2015'
    )
    parser.add_argument('-d', '--debug', action='store_true',
                        help='Switch to DEBUG log level')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Enable verbose mode')
    parser.add_argument('-i', '--interval', default=defaults.run_interval, type=float,
                        help='Polling interval')
    parser.add_argument('-m', '--memory', default=defaults.memory_trigger, type=int,
                        help='Report processes using more than n bytes')
    parser.add_argument('-c', '--cpu', default=defaults.cpu_trigger, type=float,
                        help='Report processes using more than n percent CPU')
    opts = parser.parse_args()
    setup_logger(debug=opts.debug, verbose=opts.verbose)
    pwatch(opts.interval, opts.memory, opts.cpu)


if __name__ == "__main__":
    main()



"""
Setuptools script for pwatch

"""
from setuptools import setup

def readme():
    """
    Extracts readme contents

    """
    with open('README.md') as f:
        return f.read()

def requirements():
    """
    Extacts requirements.txt contents

    """
    with open('requirements.txt') as f:
        return f.read().splitlines()

setup(
    name='pwatch',
    version='1.1.1',
    description='Simple tool to watch and report processes with excessive resource usage',
    long_description=readme(),
    license='MIT',
    url='https://github.com/quanta-computing/pwatch',
    author="Matthieu 'Korrigan' Rosinski",
    author_email='mro@quanta-computing.com',
    classifiers = [
        'License :: OSI Approved :: MIT License',
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Topic :: System :: Monitoring',
        'Topic :: Utilities',
    ],
    packages=['pwatch'],
    entry_points={
        'console_scripts': [
            'pwatch = pwatch.__main__:main',
        ],
    },
    install_requires=requirements(),
)
